import math
import random


def hfunc(text, seed=0):
    """
        Хэш-функция one_at_a_time (Jenkins)
    """
    h = 0
    for c in text:
        h += ord(c) + seed
        h += h << 10
        h ^= h >> 6
    h += h << 3
    h ^= h >> 11
    h += h << 15
    return h & 0xffffffff


class Filter1:
    def __init__(self):
        self.bits = 0

    def add(self, text):
        a = hfunc(text)
        self.bits = self.bits | a

    def contains(self, text):
        a = hfunc(text)
        if self.bits & a == a:
            return True
        return False


class Filter2:
    def __init__(self, bits_per_hash):
        self.hashs = []
        self.b_p_h = 2 ** bits_per_hash - 1

    def add(self, text):
        self.hashs.append(hfunc(text) & self.b_p_h)

    def contains(self, text):
        return hfunc(text) & self.b_p_h in self.hashs


class Filter3:
    def __init__(self, size):
        self.size = size
        self.bit_list = [0 for i in range(self.size)]

    def add(self, text):
        index = hfunc(text) % self.size
        self.bit_list[index] = 1

    def contains(self, text):
        index = hfunc(text) % self.size
        return self.bit_list[index] == 1


class Filter4:
    def __init__(self, size, nfuncs):
        self.table = [0]*size
        self.nfuncs = nfuncs

    def add(self, text):
        for seed in range(self.nfuncs):
            a = hfunc(text, seed) % len(self.table)
            self.table[a] = 1

    def contains(self, text):
        true = 0
        for seed in range(self.nfuncs):
            a = hfunc(text, seed) % len(self.table)
            if self.table[a]:
                true += 1
        if true == self.nfuncs:
            return True
        return False


class Filter5:
    def __init__(self, tmp, nfuncs=5):
        self.size = tmp * 10
        self.bit_list = [0 for i in range(self.size)]
        self.nfuncs = nfuncs

    def add(self, text):
        for i in range(self.nfuncs):
            index = hfunc(text, i) % self.size
            self.bit_list[index] = 1

    def contains(self, text):
        check = 1
        for i in range(self.nfuncs):
            index = hfunc(text, i) % self.size
            check = check & (self.bit_list[index] == 1)
        return check


def shuffle(text):
    lst = list(text)
    random.shuffle(lst)
    return "".join(lst)


def test(flt):
    print(type(flt).__name__)
    for e in EMAILS:
        flt.add(e)
    fp = 0
    for e in EMAILS:
        if not flt.contains(e):
            print("Получен ложноотрицательный результат!")
    for e in FALSE_EMAILS:
        if flt.contains(e) and e not in EMAILS:
            fp += 1
    print("Ложноположительных срабатываний: %.1f%%" % (100 * fp / len(EMAILS)))


def test_hfunc(data):
    hashes = [hfunc(x) % len(data) for x in data]
    return len(set(hashes))


if __name__ == '__main__':
    random.seed(42)
    tmp = 10000
    with open("emails.txt") as f:
        EMAILS = f.read().split('\n')[:tmp]
    FALSE_EMAILS = [shuffle(e) for e in EMAILS]

f1 = Filter1()
f2 = Filter2(10)
f3 = Filter3(1000)
f4 = Filter4(1000, 1)
f5 = Filter5(tmp)
test(f5)
